# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import sys
from ..log import logger

def running_in_gunicorn():
    """
    Determines if we are running inside of a gunicorn process.

    @return:  Boolean
    """
    process_check = False

    try:
        # Is this a gunicorn process?
        if hasattr(sys, 'argv'):
            for arg in sys.argv:
                if arg.find('gunicorn') >= 0:
                    process_check = True
        elif os.path.isfile("/proc/self/cmdline"):
            with open("/proc/self/cmdline") as cmd:
                contents = cmd.read()

            parts = contents.split('\0')
            parts.pop()
            cmdline = " ".join(parts)

            if cmdline.find('gunicorn') >= 0:
                process_check = True

        return process_check
    except Exception:
        logger.debug("Instana.log.running_in_gunicorn: ", exc_info=True)
        return False


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import sys
import time
import random

_rnd = random.Random()
_current_pid = 0

BAD_ID = "BADCAFFE"  # Bad Caffe

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    string_types = basestring
else:
    string_types = str


def generate_id():
    """ Generate a 64bit base 16 ID for use as a Span or Trace ID """
    global _current_pid

    pid = os.getpid()
    if _current_pid != pid:
        _current_pid = pid
        _rnd.seed(int(1000000 * time.time()) ^ pid)
    new_id = format(_rnd.randint(0, 18446744073709551615), '02x')

    if len(new_id) < 16:
        new_id = new_id.zfill(16)

    return new_id


def header_to_long_id(header):
    """
    We can receive headers in the following formats:
      1. unsigned base 16 hex string (or bytes) of variable length
      2. [eventual]

    :param header: the header to analyze, validate and convert (if needed)
    :return: a valid ID to be used internal to the tracer
    """
    if PY3 is True and isinstance(header, bytes):
        header = header.decode('utf-8')

    if not isinstance(header, string_types):
        return BAD_ID

    try:
        # Test that header is truly a hexadecimal value before we try to convert
        int(header, 16)

        length = len(header)
        if length < 16:
            # Left pad ID with zeros
            header = header.zfill(16)

        return header
    except ValueError:
        return BAD_ID


def header_to_id(header):
    """
    We can receive headers in the following formats:
      1. unsigned base 16 hex string (or bytes) of variable length
      2. [eventual]

    :param header: the header to analyze, validate and convert (if needed)
    :return: a valid ID to be used internal to the tracer
    """
    if PY3 is True and isinstance(header, bytes):
        header = header.decode('utf-8')

    if not isinstance(header, string_types):
        return BAD_ID

    try:
        # Test that header is truly a hexadecimal value before we try to convert
        int(header, 16)

        length = len(header)
        if length < 16:
            # Left pad ID with zeros
            header = header.zfill(16)
        elif length > 16:
            # Phase 0: Discard everything but the last 16byte
            header = header[-16:]
        return header
    except ValueError:
        return BAD_ID



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

import re
import flask

import opentracing
import opentracing.ext.tags as ext
import wrapt

from ...log import logger
from ...singletons import agent, tracer
from ...util.secrets import strip_secrets_from_query

path_tpl_re = re.compile('<.*>')


def before_request_with_instana(*argv, **kwargs):
    try:
        env = flask.request.environ
        ctx = None

        ctx = tracer.extract(opentracing.Format.HTTP_HEADERS, env)

        flask.g.scope = tracer.start_active_span('wsgi', child_of=ctx)
        span = flask.g.scope.span

        if agent.options.extra_http_headers is not None:
            for custom_header in agent.options.extra_http_headers:
                # Headers are available in this format: HTTP_X_CAPTURE_THIS
                header = ('HTTP_' + custom_header.upper()).replace('-', '_')
                if header in env:
                    span.set_tag("http.header.%s" % custom_header, env[header])

        span.set_tag(ext.HTTP_METHOD, flask.request.method)
        if 'PATH_INFO' in env:
            span.set_tag(ext.HTTP_URL, env['PATH_INFO'])
        if 'QUERY_STRING' in env and len(env['QUERY_STRING']):
            scrubbed_params = strip_secrets_from_query(env['QUERY_STRING'], agent.options.secrets_matcher,
                                                       agent.options.secrets_list)
            span.set_tag("http.params", scrubbed_params)
        if 'HTTP_HOST' in env:
            span.set_tag("http.host", env['HTTP_HOST'])

        if hasattr(flask.request.url_rule, 'rule') and \
                path_tpl_re.search(flask.request.url_rule.rule) is not None:
            path_tpl = flask.request.url_rule.rule.replace("<", "{")
            path_tpl = path_tpl.replace(">", "}")
            span.set_tag("http.path_tpl", path_tpl)
    except:
        logger.debug("Flask before_request", exc_info=True)

    return None


def after_request_with_instana(response):
    scope = None
    try:
        # If we're not tracing, just return
        if not hasattr(flask.g, 'scope'):
            return response

        scope = flask.g.scope
        if scope is not None:
            span = scope.span

            if 500 <= response.status_code <= 511:
                span.mark_as_errored()

            span.set_tag(ext.HTTP_STATUS_CODE, int(response.status_code))
            tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, response.headers)
            response.headers.add('Server-Timing', "intid;desc=%s" % scope.span.context.trace_id)
    except:
        logger.debug("Flask after_request", exc_info=True)
    finally:
        if scope is not None:
            scope.close()
            flask.g.scope = None
    return response


def teardown_request_with_instana(*argv, **kwargs):
    """
    In the case of exceptions, after_request_with_instana isn't called
    so we capture those cases here.
    """
    if hasattr(flask.g, 'scope') and flask.g.scope is not None:
        if len(argv) > 0 and argv[0] is not None:
            scope = flask.g.scope
            scope.span.log_exception(argv[0])
            if ext.HTTP_STATUS_CODE not in scope.span.tags:
                scope.span.set_tag(ext.HTTP_STATUS_CODE, 500)
        flask.g.scope.close()
        flask.g.scope = None


@wrapt.patch_function_wrapper('flask', 'Flask.full_dispatch_request')
def full_dispatch_request_with_instana(wrapped, instance, argv, kwargs):
    if not hasattr(instance, '_stan_wuz_here'):
        logger.debug("Flask(vanilla): Applying flask before/after instrumentation funcs")
        setattr(instance, "_stan_wuz_here", True)
        instance.after_request(after_request_with_instana)
        instance.before_request(before_request_with_instana)
        instance.teardown_request(teardown_request_with_instana)
    return wrapped(*argv, **kwargs)


logger.debug("Instrumenting flask (without blinker support)")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import re

try:
    # Python 3
    from urllib.parse import unquote
except ImportError:
    # Python 2
    from urllib import unquote

# _storage_api defines a conversion of Google Storage JSON API requests into span tags as follows:
#   request_method -> path_matcher -> collector
#
# * request method - the HTTP method used to make an API request (GET, POST, etc.)
# * path_matcher - either a string or a regex applied to the API request path (string values match first).
# * collector - a lambda returning a dict of span from  API request query string.
#               parameters and request body data. If a regex is used as a path matcher, the match result
#               will be provided as a third argument.
#
# The API documentation can be found at https://cloud.google.com/storage/docs/json_api
_storage_api = {
    'GET': {
        #####################
        # Bucket operations #
        #####################
        '/b': lambda params, data: {
            'gcs.op': 'buckets.list',
            'gcs.projectId': params.get('project', None)
        },
        re.compile('^/b/(?P<bucket>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'buckets.get',
            'gcs.bucket': unquote(match.group('bucket')),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/iam$'): lambda params, data, match: {
            'gcs.op': 'buckets.getIamPolicy',
            'gcs.bucket': unquote(match.group('bucket')),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/iam/testPermissions$'): lambda params, data, match: {
            'gcs.op': 'buckets.testIamPermissions',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        ##########################
        # Object/blob operations #
        ##########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)$'): lambda params, data, match: {
            'gcs.op': params.get('alt', 'json') == 'media' and 'objects.get' or 'objects.attrs',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/o$'): lambda params, data, match: {
            'gcs.op': 'objects.list',
            'gcs.bucket': unquote(match.group('bucket'))
        },

        ##################################
        # Default object ACLs operations #
        ##################################
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.get',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.entity': unquote(match.group('entity'))
        },
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.list',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        #########################
        # Object ACL operations #
        #########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.get',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
            'gcs.entity': unquote(match.group('entity'))
        },
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.list',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object'))
        },

        ########################
        # HMAC keys operations #
        ########################
        re.compile('^/projects/(?P<project>[^/]+)/hmacKeys$'): lambda params, data, match: {
            'gcs.op': 'hmacKeys.list',
            'gcs.projectId': unquote(match.group('project'))
        },
        re.compile('^/projects/(?P<project>[^/]+)/hmacKeys/(?P<accessId>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'hmacKeys.get',
            'gcs.projectId': unquote(match.group('project')),
            'gcs.accessId': unquote(match.group('accessId'))
        },

        ##############################
        # Service account operations #
        ##############################
        re.compile('^/projects/(?P<project>[^/]+)/serviceAccount$'): lambda params, data, match: {
            'gcs.op': 'serviceAccount.get',
            'gcs.projectId': unquote(match.group('project'))
        }
    },
    'POST': {
        #####################
        # Bucket operations #
        #####################
        '/b': lambda params, data: {
            'gcs.op': 'buckets.insert',
            'gcs.projectId': params.get('project', None),
            'gcs.bucket': data.get('name', None),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/lockRetentionPolicy$'): lambda params, data, match: {
            'gcs.op': 'buckets.lockRetentionPolicy',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        ##########################
        # Object/blob operations #
        ##########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/compose$'): lambda params, data, match: {
            'gcs.op': 'objects.compose',
            'gcs.destinationBucket': unquote(match.group('bucket')),
            'gcs.destinationObject': unquote(match.group('object')),
            'gcs.sourceObjects': ','.join(
                ['%s/%s' % (unquote(match.group('bucket')), o['name']) for o in data.get('sourceObjects', []) if 'name' in o]
            )
        },
        re.compile('^/b/(?P<srcBucket>[^/]+)/o/(?P<srcObject>[^/]+)/copyTo/b/(?P<destBucket>[^/]+)/o/(?P<destObject>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objects.copy',
            'gcs.destinationBucket': unquote(match.group('destBucket')),
            'gcs.destinationObject': unquote(match.group('destObject')),
            'gcs.sourceBucket': unquote(match.group('srcBucket')),
            'gcs.sourceObject': unquote(match.group('srcObject')),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/o$'): lambda params, data, match: {
            'gcs.op': 'objects.insert',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': params.get('name', data.get('name', None)),
        },
        re.compile('^/b/(?P<srcBucket>[^/]+)/o/(?P<srcObject>[^/]+)/rewriteTo/b/(?P<destBucket>[^/]+)/o/(?P<destObject>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objects.rewrite',
            'gcs.destinationBucket': unquote(match.group('destBucket')),
            'gcs.destinationObject': unquote(match.group('destObject')),
            'gcs.sourceBucket': unquote(match.group('srcBucket')),
            'gcs.sourceObject': unquote(match.group('srcObject')),
        },

        ######################
        # Channel operations #
        ######################
        '/channels/stop': lambda params, data: {
            'gcs.op': 'channels.stop',
            'gcs.entity': data.get('id', None)
        },

        ##################################
        # Default object ACLs operations #
        ##################################
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.insert',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.entity': data.get('entity', None)
        },

        #########################
        # Object ACL operations #
        #########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.insert',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
            'gcs.entity': data.get('entity', None)
        },

        ########################
        # HMAC keys operations #
        ########################
        re.compile('^/projects/(?P<project>[^/]+)/hmacKeys$'): lambda params, data, match: {
            'gcs.op': 'hmacKeys.create',
            'gcs.projectId': unquote(match.group('project'))
        }
    },
    'PATCH': {
        #####################
        # Bucket operations #
        #####################
        re.compile('^/b/(?P<bucket>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'buckets.patch',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        ##########################
        # Object/blob operations #
        ##########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objects.patch',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
        },

        ##################################
        # Default object ACLs operations #
        ##################################
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.patch',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.entity': unquote(match.group('entity'))
        },

        #########################
        # Object ACL operations #
        #########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.patch',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
            'gcs.entity': unquote(match.group('entity'))
        }
    },
    'PUT': {
        #####################
        # Bucket operations #
        #####################
        re.compile('^/b/(?P<bucket>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'buckets.update',
            'gcs.bucket': unquote(match.group('bucket')),
        },
        re.compile('^/b/(?P<bucket>[^/]+)/iam$'): lambda params, data, match: {
            'gcs.op': 'buckets.setIamPolicy',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        ##########################
        # Object/blob operations #
        ##########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objects.update',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
        },

        ##################################
        # Default object ACLs operations #
        ##################################
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.update',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.entity': unquote(match.group('entity'))
        },

        #########################
        # Object ACL operations #
        #########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.update',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
            'gcs.entity': unquote(match.group('entity'))
        },

        ########################
        # HMAC keys operations #
        ########################
        re.compile('^/projects/(?P<project>[^/]+)/hmacKeys/(?P<accessId>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'hmacKeys.update',
            'gcs.projectId': unquote(match.group('project')),
            'gcs.accessId': unquote(match.group('accessId'))
        }
    },
    'DELETE': {
        #####################
        # Bucket operations #
        #####################
        re.compile('^/b/(?P<bucket>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'buckets.delete',
            'gcs.bucket': unquote(match.group('bucket')),
        },

        ##########################
        # Object/blob operations #
        ##########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objects.delete',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
        },

        ##################################
        # Default object ACLs operations #
        ##################################
        re.compile('^/b/(?P<bucket>[^/]+)/defaultObjectAcl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'defaultAcls.delete',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.entity': unquote(match.group('entity'))
        },

        #########################
        # Object ACL operations #
        #########################
        re.compile('^/b/(?P<bucket>[^/]+)/o/(?P<object>[^/]+)/acl/(?P<entity>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'objectAcls.delete',
            'gcs.bucket': unquote(match.group('bucket')),
            'gcs.object': unquote(match.group('object')),
            'gcs.entity': unquote(match.group('entity'))
        },

        ########################
        # HMAC keys operations #
        ########################
        re.compile('^/projects/(?P<project>[^/]+)/hmacKeys/(?P<accessId>[^/]+)$'): lambda params, data, match: {
            'gcs.op': 'hmacKeys.delete',
            'gcs.projectId': unquote(match.group('project')),
            'gcs.accessId': unquote(match.group('accessId'))
        }
    }
}



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
from .app import pyramid_server as server
from ..utils import launch_background_thread

app_thread = None

if 'CASSANDRA_TEST' not in os.environ:
    app_thread = launch_background_thread(server.serve_forever, "Pyramid")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import time
from celery import shared_task
from instana.singletons import tracer
from ..helpers import get_first_span_by_filter

# TODO: Refactor to class based tests

@shared_task
def add(x, y):
    return x + y


@shared_task
def will_raise_error():
    raise Exception('This is a simulated error')


def filter_out_ping_tasks(spans):
    filtered_spans = []
    for span in spans:
        is_ping_task = (span.n == 'celery-worker' and span.data['celery']['task'] == 'celery.ping')
        if not is_ping_task:
            filtered_spans.append(span)
    return filtered_spans


def setup_method():
    """ Clear all spans before a test run """
    tracer.recorder.clear_spans()


def test_apply_async(celery_app, celery_worker):
    result = None
    with tracer.start_active_span('test'):
        result = add.apply_async(args=(4, 5))

    # Wait for jobs to finish
    time.sleep(0.5)

    spans = filter_out_ping_tasks(tracer.recorder.queued_spans())
    assert len(spans) == 3

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "celery-client"
    client_span = get_first_span_by_filter(spans, filter)
    assert(client_span)

    filter = lambda span: span.n == "celery-worker"
    worker_span = get_first_span_by_filter(spans, filter)
    assert(worker_span)

    assert(client_span.t == test_span.t)
    assert(client_span.t == worker_span.t)
    assert(client_span.p == test_span.s)

    assert("tests.frameworks.test_celery.add" == client_span.data["celery"]["task"])
    assert("redis" == client_span.data["celery"]["scheme"])
    assert("localhost" == client_span.data["celery"]["host"])
    assert("6379" == client_span.data["celery"]["port"])
    assert(client_span.data["celery"]["task_id"])
    assert(client_span.data["celery"]["error"] == None)
    assert(client_span.ec == None)

    assert("tests.frameworks.test_celery.add" == worker_span.data["celery"]["task"])
    assert("redis" == worker_span.data["celery"]["scheme"])
    assert("localhost" == worker_span.data["celery"]["host"])
    assert("6379" == worker_span.data["celery"]["port"])
    assert(worker_span.data["celery"]["task_id"])
    assert(worker_span.data["celery"]["error"] == None)
    assert(worker_span.data["celery"]["retry-reason"] == None)
    assert(worker_span.ec == None)


def test_delay(celery_app, celery_worker):
    result = None
    with tracer.start_active_span('test'):
        result = add.delay(4, 5)

    # Wait for jobs to finish
    time.sleep(0.5)

    spans = filter_out_ping_tasks(tracer.recorder.queued_spans())
    assert len(spans) == 3

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "celery-client"
    client_span = get_first_span_by_filter(spans, filter)
    assert(client_span)

    filter = lambda span: span.n == "celery-worker"
    worker_span = get_first_span_by_filter(spans, filter)
    assert(worker_span)

    assert(client_span.t == test_span.t)
    assert(client_span.t == worker_span.t)
    assert(client_span.p == test_span.s)

    assert("tests.frameworks.test_celery.add" == client_span.data["celery"]["task"])
    assert("redis" == client_span.data["celery"]["scheme"])
    assert("localhost" == client_span.data["celery"]["host"])
    assert("6379" == client_span.data["celery"]["port"])
    assert(client_span.data["celery"]["task_id"])
    assert(client_span.data["celery"]["error"] == None)
    assert(client_span.ec == None)

    assert("tests.frameworks.test_celery.add" == worker_span.data["celery"]["task"])
    assert("redis" == worker_span.data["celery"]["scheme"])
    assert("localhost" == worker_span.data["celery"]["host"])
    assert("6379" == worker_span.data["celery"]["port"])
    assert(worker_span.data["celery"]["task_id"])
    assert(worker_span.data["celery"]["error"] == None)
    assert(worker_span.data["celery"]["retry-reason"] == None)
    assert(worker_span.ec == None)


def test_send_task(celery_app, celery_worker):
    result = None
    with tracer.start_active_span('test'):
        result = celery_app.send_task('tests.frameworks.test_celery.add', (1, 2))

    # Wait for jobs to finish
    time.sleep(0.5)

    spans = filter_out_ping_tasks(tracer.recorder.queued_spans())
    assert len(spans) == 3

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "celery-client"
    client_span = get_first_span_by_filter(spans, filter)
    assert(client_span)

    filter = lambda span: span.n == "celery-worker"
    worker_span = get_first_span_by_filter(spans, filter)
    assert(worker_span)

    assert(client_span.t == test_span.t)
    assert(client_span.t == worker_span.t)
    assert(client_span.p == test_span.s)

    assert("tests.frameworks.test_celery.add" == client_span.data["celery"]["task"])
    assert("redis" == client_span.data["celery"]["scheme"])
    assert("localhost" == client_span.data["celery"]["host"])
    assert("6379" == client_span.data["celery"]["port"])
    assert(client_span.data["celery"]["task_id"])
    assert(client_span.data["celery"]["error"] == None)
    assert(client_span.ec == None)

    assert("tests.frameworks.test_celery.add" == worker_span.data["celery"]["task"])
    assert("redis" == worker_span.data["celery"]["scheme"])
    assert("localhost" == worker_span.data["celery"]["host"])
    assert("6379" == worker_span.data["celery"]["port"])
    assert(worker_span.data["celery"]["task_id"])
    assert(worker_span.data["celery"]["error"] == None)
    assert(worker_span.data["celery"]["retry-reason"] == None)
    assert(worker_span.ec == None)


def test_error_reporting(celery_app, celery_worker):
    result = None
    with tracer.start_active_span('test'):
        result = will_raise_error.apply_async()

    # Wait for jobs to finish
    time.sleep(0.5)

    spans = filter_out_ping_tasks(tracer.recorder.queued_spans())
    assert len(spans) == 4

    filter = lambda span: span.n == "sdk"
    test_span = get_first_span_by_filter(spans, filter)
    assert(test_span)

    filter = lambda span: span.n == "celery-client"
    client_span = get_first_span_by_filter(spans, filter)
    assert(client_span)

    filter = lambda span: span.n == "log"
    log_span = get_first_span_by_filter(spans, filter)
    assert(log_span)

    filter = lambda span: span.n == "celery-worker"
    worker_span = get_first_span_by_filter(spans, filter)
    assert(worker_span)

    assert(client_span.t == test_span.t)
    assert(client_span.t == worker_span.t)
    assert(client_span.t == log_span.t)

    assert(client_span.p == test_span.s)
    assert(worker_span.p == client_span.s)
    assert(log_span.p == worker_span.s)

    assert("tests.frameworks.test_celery.will_raise_error" == client_span.data["celery"]["task"])
    assert("redis" == client_span.data["celery"]["scheme"])
    assert("localhost" == client_span.data["celery"]["host"])
    assert("6379" == client_span.data["celery"]["port"])
    assert(client_span.data["celery"]["task_id"])
    assert(client_span.data["celery"]["error"] == None)
    assert(client_span.ec == None)

    assert("tests.frameworks.test_celery.will_raise_error" == worker_span.data["celery"]["task"])
    assert("redis" == worker_span.data["celery"]["scheme"])
    assert("localhost" == worker_span.data["celery"]["host"])
    assert("6379" == worker_span.data["celery"]["port"])
    assert(worker_span.data["celery"]["task_id"])
    assert(worker_span.data["celery"]["error"] == 'This is a simulated error')
    assert(worker_span.data["celery"]["retry-reason"] == None)
    assert(worker_span.ec == 1)




# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from __future__ import absolute_import

import os
import pika
import unittest
import mock
import threading
import time

from ..helpers import testenv
from instana.singletons import tracer


class _TestPika(unittest.TestCase):
    @staticmethod
    @mock.patch('pika.connection.Connection')
    def _create_connection(connection_class_mock=None):
        return connection_class_mock()

    def _create_obj(self):
        raise NotImplementedError()

    def setUp(self):
        self.recorder = tracer.recorder
        self.recorder.clear_spans()

        self.connection = self._create_connection()
        self._on_openok_callback = mock.Mock()
        self.obj = self._create_obj()

    def tearDown(self):
        del self.connection
        del self._on_openok_callback
        del self.obj


class TestPikaChannel(_TestPika):
    def _create_obj(self):
        return pika.channel.Channel(self.connection, 1, self._on_openok_callback)

    @mock.patch('pika.spec.Basic.Publish')
    @mock.patch('pika.channel.Channel._send_method')
    def test_basic_publish(self, send_method, _unused):
        self.obj._set_state(self.obj.OPEN)

        with tracer.start_active_span("testing"):
            self.obj.basic_publish("test.exchange", "test.queue", "Hello!")

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        rabbitmq_span = spans[0]
        test_span = spans[1]

        self.assertIsNone(tracer.active_span)

        # Same traceId
        self.assertEqual(test_span.t, rabbitmq_span.t)

        # Parent relationships
        self.assertEqual(rabbitmq_span.p, test_span.s)

        # Error logging
        self.assertIsNone(test_span.ec)
        self.assertIsNone(rabbitmq_span.ec)

        # Span tags
        self.assertEqual("test.exchange", rabbitmq_span.data["rabbitmq"]["exchange"])
        self.assertEqual('publish', rabbitmq_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(rabbitmq_span.data["rabbitmq"]["address"])
        self.assertEqual("test.queue", rabbitmq_span.data["rabbitmq"]["key"])
        self.assertIsNotNone(rabbitmq_span.stack)
        self.assertTrue(type(rabbitmq_span.stack) is list)
        self.assertGreater(len(rabbitmq_span.stack), 0)

        send_method.assert_called_once_with(
            pika.spec.Basic.Publish(
                exchange="test.exchange",
                routing_key="test.queue"), (pika.spec.BasicProperties(headers={
                "X-INSTANA-T": rabbitmq_span.t,
                "X-INSTANA-S": rabbitmq_span.s,
                "X-INSTANA-L": "1"
            }), b"Hello!"))

    @mock.patch('pika.spec.Basic.Publish')
    @mock.patch('pika.channel.Channel._send_method')
    def test_basic_publish_with_headers(self, send_method, _unused):
        self.obj._set_state(self.obj.OPEN)

        with tracer.start_active_span("testing"):
            self.obj.basic_publish("test.exchange",
                                   "test.queue",
                                   "Hello!",
                                   pika.BasicProperties(headers={
                                       "X-Custom-1": "test"
                                   }))

        spans = self.recorder.queued_spans()
        self.assertEqual(2, len(spans))

        rabbitmq_span = spans[0]
        test_span = spans[1]

        send_method.assert_called_once_with(
            pika.spec.Basic.Publish(
                exchange="test.exchange",
                routing_key="test.queue"), (pika.spec.BasicProperties(headers={
                "X-Custom-1": "test",
                "X-INSTANA-T": rabbitmq_span.t,
                "X-INSTANA-S": rabbitmq_span.s,
                "X-INSTANA-L": "1"
            }), b"Hello!"))

    @mock.patch('pika.spec.Basic.Get')
    def test_basic_get(self, _unused):
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties()

        method_frame = pika.frame.Method(1, pika.spec.Basic.GetOk)
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_get("test.queue", cb)
        self.obj._on_getok(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.t)
        self.assertIsNone(rabbitmq_span.p)
        self.assertIsNotNone(rabbitmq_span.s)

        # Error logging
        self.assertIsNone(rabbitmq_span.ec)

        # Span tags
        self.assertIsNone(rabbitmq_span.data["rabbitmq"]["exchange"])
        self.assertEqual("consume", rabbitmq_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(rabbitmq_span.data["rabbitmq"]["address"])
        self.assertEqual("test.queue", rabbitmq_span.data["rabbitmq"]["queue"])
        self.assertIsNotNone(rabbitmq_span.stack)
        self.assertTrue(type(rabbitmq_span.stack) is list)
        self.assertGreater(len(rabbitmq_span.stack), 0)

        cb.assert_called_once_with(self.obj, pika.spec.Basic.GetOk, properties, body)

    @mock.patch('pika.spec.Basic.Get')
    def test_basic_get_with_trace_context(self, _unused):
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties(headers={
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1"
        })

        method_frame = pika.frame.Method(1, pika.spec.Basic.GetOk)
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_get("test.queue", cb)
        self.obj._on_getok(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # Trace context propagation
        self.assertEqual("0000000000000001", rabbitmq_span.t)
        self.assertEqual("0000000000000002", rabbitmq_span.p)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.s)
        self.assertNotEqual(rabbitmq_span.p, rabbitmq_span.s)

    @mock.patch('pika.spec.Basic.Consume')
    def test_basic_consume(self, _unused):
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties()

        method_frame = pika.frame.Method(1, pika.spec.Basic.Deliver(consumer_tag="test"))
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_consume("test.queue", cb, consumer_tag="test")
        self.obj._on_deliver(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.t)
        self.assertIsNone(rabbitmq_span.p)
        self.assertIsNotNone(rabbitmq_span.s)

        # Error logging
        self.assertIsNone(rabbitmq_span.ec)

        # Span tags
        self.assertIsNone(rabbitmq_span.data["rabbitmq"]["exchange"])
        self.assertEqual("consume", rabbitmq_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(rabbitmq_span.data["rabbitmq"]["address"])
        self.assertEqual("test.queue", rabbitmq_span.data["rabbitmq"]["queue"])
        self.assertIsNotNone(rabbitmq_span.stack)
        self.assertTrue(type(rabbitmq_span.stack) is list)
        self.assertGreater(len(rabbitmq_span.stack), 0)

        cb.assert_called_once_with(self.obj, method_frame.method, properties, body)

    @mock.patch('pika.spec.Basic.Consume')
    def test_basic_consume_with_trace_context(self, _unused):
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties(headers={
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1"
        })

        method_frame = pika.frame.Method(1, pika.spec.Basic.Deliver(consumer_tag="test"))
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_consume("test.queue", cb, consumer_tag="test")
        self.obj._on_deliver(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # Trace context propagation
        self.assertEqual("0000000000000001", rabbitmq_span.t)
        self.assertEqual("0000000000000002", rabbitmq_span.p)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.s)
        self.assertNotEqual(rabbitmq_span.p, rabbitmq_span.s)


class TestPikaBlockingChannel(_TestPika):
    @mock.patch('pika.channel.Channel', spec=pika.channel.Channel)
    def _create_obj(self, channel_impl):
        self.impl = channel_impl()
        self.impl.channel_number = 1

        return pika.adapters.blocking_connection.BlockingChannel(self.impl, self.connection)

    def _generate_delivery(self, consumer_tag, properties, body):
        from pika.adapters.blocking_connection import _ConsumerDeliveryEvt

        # Wait until queue consumer is initialized
        while self.obj._queue_consumer_generator is None:
            time.sleep(0.25)

        method = pika.spec.Basic.Deliver(consumer_tag=consumer_tag)
        self.obj._on_consumer_generator_event(_ConsumerDeliveryEvt(method, properties, body))

    def test_consume(self):
        consumed_deliveries = []

        def __consume():
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)

                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        self._generate_delivery(consumer_tag, pika.BasicProperties(), "Hello!")

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.t)
        self.assertIsNone(rabbitmq_span.p)
        self.assertIsNotNone(rabbitmq_span.s)

        # Error logging
        self.assertIsNone(rabbitmq_span.ec)

        # Span tags
        self.assertIsNone(rabbitmq_span.data["rabbitmq"]["exchange"])
        self.assertEqual("consume", rabbitmq_span.data["rabbitmq"]["sort"])
        self.assertIsNotNone(rabbitmq_span.data["rabbitmq"]["address"])
        self.assertEqual("test.queue", rabbitmq_span.data["rabbitmq"]["queue"])
        self.assertIsNotNone(rabbitmq_span.stack)
        self.assertTrue(type(rabbitmq_span.stack) is list)
        self.assertGreater(len(rabbitmq_span.stack), 0)

        self.assertEqual(1, len(consumed_deliveries))

    def test_consume_with_trace_context(self):
        consumed_deliveries = []

        def __consume():
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)

                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        self._generate_delivery(consumer_tag, pika.BasicProperties(headers={
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1"
        }), "Hello!")

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        self.assertEqual(1, len(spans))

        rabbitmq_span = spans[0]

        self.assertIsNone(tracer.active_span)

        # Trace context propagation
        self.assertEqual("0000000000000001", rabbitmq_span.t)
        self.assertEqual("0000000000000002", rabbitmq_span.p)

        # A new span has been started
        self.assertIsNotNone(rabbitmq_span.s)
        self.assertNotEqual(rabbitmq_span.p, rabbitmq_span.s)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

"""
Instrumentation for Sanic
https://sanicframework.org/en/
"""
try:
    import sanic
    import wrapt
    import opentracing
    from ..log import logger
    from ..singletons import async_tracer, agent
    from ..util.secrets import strip_secrets_from_query
    from ..util.traceutils import extract_custom_headers


    @wrapt.patch_function_wrapper('sanic.exceptions', 'SanicException.__init__')
    def exception_with_instana(wrapped, instance, args, kwargs):
        try:
            message = kwargs.get("message") or args[0]
            status_code = kwargs.get("status_code")
            span = async_tracer.active_span

            if all([span, status_code, message]) and (500 <= status_code <= 599):
                span.set_tag("http.error", message)
                try:
                    wrapped(*args, **kwargs)
                except Exception as exc:
                    span.log_exception(exc)
            else:
                wrapped(*args, **kwargs)
        except Exception:
            logger.debug("exception_with_instana: ", exc_info=True)
            wrapped(*args, **kwargs)


    def response_details(span, response):
        try:
            status_code = response.status
            if status_code is not None:
                if 500 <= int(status_code) <= 511:
                    span.mark_as_errored()
                span.set_tag('http.status_code', status_code)

            if response.headers is not None:
                async_tracer.inject(span.context, opentracing.Format.HTTP_HEADERS, response.headers)
            response.headers['Server-Timing'] = "intid;desc=%s" % span.context.trace_id
        except Exception:
            logger.debug("send_wrapper: ", exc_info=True)


    if hasattr(sanic.response.BaseHTTPResponse, "send"):
        @wrapt.patch_function_wrapper('sanic.response', 'BaseHTTPResponse.send')
        async def send_with_instana(wrapped, instance, args, kwargs):
            span = async_tracer.active_span
            if span is None:
                await wrapped(*args, **kwargs)
            else:
                response_details(span=span, response=instance)
                try:
                    await wrapped(*args, **kwargs)
                except Exception as exc:
                    span.log_exception(exc)
                    raise
    else:
        @wrapt.patch_function_wrapper('sanic.server', 'HttpProtocol.write_response')
        def write_with_instana(wrapped, instance, args, kwargs):
            response = args[0]
            span = async_tracer.active_span
            if span is None:
                wrapped(*args, **kwargs)
            else:
                response_details(span=span, response=response)
                try:
                    wrapped(*args, **kwargs)
                except Exception as exc:
                    span.log_exception(exc)
                    raise


        @wrapt.patch_function_wrapper('sanic.server', 'HttpProtocol.stream_response')
        async def stream_with_instana(wrapped, instance, args, kwargs):
            response = args[0]
            span = async_tracer.active_span
            if span is None:
                await wrapped(*args, **kwargs)
            else:
                response_details(span=span, response=response)
                try:
                    await wrapped(*args, **kwargs)
                except Exception as exc:
                    span.log_exception(exc)
                    raise


    @wrapt.patch_function_wrapper('sanic.app', 'Sanic.handle_request')
    async def handle_request_with_instana(wrapped, instance, args, kwargs):

        try:
            request = args[0]
            try:  # scheme attribute is calculated in the sanic handle_request method for v19, not yet present
                if "http" not in request.scheme:
                    return await wrapped(*args, **kwargs)
            except AttributeError:
                pass
            headers = request.headers.copy()
            ctx = async_tracer.extract(opentracing.Format.HTTP_HEADERS, headers)
            with async_tracer.start_active_span("asgi", child_of=ctx) as scope:
                scope.span.set_tag('span.kind', 'entry')
                scope.span.set_tag('http.path', request.path)
                scope.span.set_tag('http.method', request.method)
                scope.span.set_tag('http.host', request.host)
                if hasattr(request, "url"):
                    scope.span.set_tag("http.url", request.url)

                query = request.query_string

                if isinstance(query, (str, bytes)) and len(query):
                    if isinstance(query, bytes):
                        query = query.decode('utf-8')
                    scrubbed_params = strip_secrets_from_query(query, agent.options.secrets_matcher,
                                                               agent.options.secrets_list)
                    scope.span.set_tag("http.params", scrubbed_params)

                if agent.options.extra_http_headers is not None:
                    extract_custom_headers(scope, headers)
                await wrapped(*args, **kwargs)
                if hasattr(request, "uri_template") and request.uri_template:
                    scope.span.set_tag("http.path_tpl", request.uri_template)
                if hasattr(request, "ctx"):  # ctx attribute added in the latest v19 versions
                    request.ctx.iscope = scope
        except Exception as e:
            logger.debug("Sanic framework @ handle_request", exc_info=True)
            return await wrapped(*args, **kwargs)


    logger.debug("Instrumenting Sanic")

except ImportError:
    pass
except AttributeError:
    logger.debug("Not supported Sanic version")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

"""
Google Cloud Run Collector: Manages the periodic collection of metrics & snapshot data
"""
import os
from time import time
import requests

from instana.log import logger
from instana.collector.base import BaseCollector
from instana.util import DictionaryOfStan, validate_url
from instana.collector.helpers.google_cloud_run.process import GCRProcessHelper
from instana.collector.helpers.google_cloud_run.instance_entity import InstanceEntityHelper


class GCRCollector(BaseCollector):
    """ Collector for Google Cloud Run """

    def __init__(self, agent, service, configuration, revision):
        super(GCRCollector, self).__init__(agent)
        logger.debug("Loading Google Cloud Run Collector")

        # Indicates if this Collector has all requirements to run successfully
        self.ready_to_start = True

        self.revision = revision
        self.service = service
        self.configuration = configuration
        # Prepare the URLS that we will collect data from
        self._gcr_md_uri = os.environ.get("GOOGLE_CLOUD_RUN_METADATA_ENDPOINT", "http://metadata.google.internal")

        if self._gcr_md_uri == "" or validate_url(self._gcr_md_uri) is False:
            logger.warning("GCRCollector: GOOGLE_CLOUD_RUN_METADATA_ENDPOINT not in environment or invalid URL.  "
                           "Instana will not be able to monitor this environment")
            self.ready_to_start = False

        self._gcr_md_project_uri = self._gcr_md_uri + '/computeMetadata/v1/project/?recursive=true'
        self._gcr_md_instance_uri = self._gcr_md_uri + '/computeMetadata/v1/instance/?recursive=true'

        # Timestamp in seconds of the last time we fetched all GCR metadata
        self.__last_gcr_md_full_fetch = 0

        # How often to do a full fetch of GCR metadata
        self.__gcr_md_full_fetch_interval = 300

        # HTTP client with keep-alive
        self._http_client = requests.Session()

        # The fully qualified ARN for this process
        self._gcp_arn = None

        # Response from the last call to
        # Instance URI
        self.instance_metadata = None

        # Response from the last call to
        # Project URI
        self.project_metadata = None

        # Populate the collection helpers
        self.helpers.append(GCRProcessHelper(self))
        self.helpers.append(InstanceEntityHelper(self))

    def start(self):
        if self.ready_to_start is False:
            logger.warning("Google Cloud Run Collector is missing requirements and cannot monitor this environment.")
            return

        super(GCRCollector, self).start()

    def __get_project_instance_metadata(self):
        """
        Get the latest data from the service revision instance entity metadata and store in the class
        @return: Boolean
        """
        try:
            # Refetch the GCR snapshot data
            self.__last_gcr_md_full_fetch = int(time())
            headers = {"Metadata-Flavor": "Google"}
            # Response from the last call to
            # ${GOOGLE_CLOUD_RUN_METADATA_ENDPOINT}/computeMetadata/v1/project/?recursive=true
            self.project_metadata = self._http_client.get(self._gcr_md_project_uri, timeout=1,
                                                          headers=headers).json()

            # Response from the last call to
            # ${GOOGLE_CLOUD_RUN_METADATA_ENDPOINT}/computeMetadata/v1/instance/?recursive=true
            self.instance_metadata = self._http_client.get(self._gcr_md_instance_uri, timeout=1,
                                                           headers=headers).json()
        except Exception:
            logger.debug("GoogleCloudRunCollector.get_project_instance_metadata", exc_info=True)

    def should_send_snapshot_data(self):
        return int(time()) - self.snapshot_data_last_sent > self.snapshot_data_interval

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = []
        payload["metrics"]["plugins"] = []

        try:

            if not self.span_queue.empty():
                payload["spans"] = self.queued_spans()

            self.fetching_start_time = int(time())
            delta = self.fetching_start_time - self.__last_gcr_md_full_fetch
            if delta < self.__gcr_md_full_fetch_interval:
                return payload

            with_snapshot = self.should_send_snapshot_data()

            # Fetch the latest metrics
            self.__get_project_instance_metadata()
            if self.instance_metadata is None and self.project_metadata is None:
                return payload

            plugins = []
            for helper in self.helpers:
                plugins.extend(
                    helper.collect_metrics(with_snapshot=with_snapshot, instance_metadata=self.instance_metadata,
                                           project_metadata=self.project_metadata))

            payload["metrics"]["plugins"] = plugins

            if with_snapshot:
                self.snapshot_data_last_sent = int(time())
        except Exception:
            logger.debug("collect_snapshot error", exc_info=True)

        return payload

    def get_instance_id(self):
        try:
            if self.instance_metadata:
                return self.instance_metadata.get("id")
        except Exception:
            logger.debug("get_instance_id error", exc_info=True)
        return None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

from __future__ import absolute_import

import wrapt
import opentracing

from ..log import logger
from ..singletons import tracer

try:
    import grpc
    from grpc._channel import _UnaryUnaryMultiCallable, _StreamUnaryMultiCallable, \
        _UnaryStreamMultiCallable, _StreamStreamMultiCallable

    SUPPORTED_TYPES = [_UnaryUnaryMultiCallable,
                       _StreamUnaryMultiCallable,
                       _UnaryStreamMultiCallable,
                       _StreamStreamMultiCallable]


    def collect_tags(span, instance, argv, kwargs):
        try:
            span.set_tag('rpc.flavor', 'grpc')

            if type(instance) in SUPPORTED_TYPES:
                method = instance._method.decode()
                target = instance._channel.target().decode()
            elif type(argv[0]) is grpc._cython.cygrpc.RequestCallEvent:
                method = argv[0].call_details.method.decode()
                target = argv[0].call_details.host.decode()
            elif len(argv) > 2:
                method = argv[2][2][1]._method.decode()
                target = argv[2][2][1]._channel.target().decode()

            span.set_tag('rpc.call', method)

            parts = target.split(':')
            if len(parts) == 2:
                span.set_tag('rpc.host', parts[0])
                span.set_tag('rpc.port', parts[1])
        except:
            logger.debug("grpc.collect_tags non-fatal error", exc_info=True)
        return span


    @wrapt.patch_function_wrapper('grpc._channel', '_UnaryUnaryMultiCallable.with_call')
    def unary_unary_with_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'unary')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_UnaryUnaryMultiCallable.future')
    def unary_unary_future_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if "metadata" not in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'unary')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_UnaryUnaryMultiCallable.__call__')
    def unary_unary_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'unary')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_StreamUnaryMultiCallable.__call__')
    def stream_unary_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'stream')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_StreamUnaryMultiCallable.with_call')
    def stream_unary_with_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'stream')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_StreamUnaryMultiCallable.future')
    def stream_unary_future_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'stream')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_UnaryStreamMultiCallable.__call__')
    def unary_stream_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'stream')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._channel', '_StreamStreamMultiCallable.__call__')
    def stream_stream_call_with_instana(wrapped, instance, argv, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None:
            return wrapped(*argv, **kwargs)

        with tracer.start_active_span("rpc-client", child_of=parent_span) as scope:
            try:
                if not "metadata" in kwargs:
                    kwargs["metadata"] = []

                kwargs["metadata"] = tracer.inject(scope.span.context, opentracing.Format.BINARY, kwargs['metadata'],
                                                   disable_w3c_trace_context=True)
                collect_tags(scope.span, instance, argv, kwargs)
                scope.span.set_tag('rpc.call_type', 'stream')

                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    @wrapt.patch_function_wrapper('grpc._server', '_call_behavior')
    def call_behavior_with_instana(wrapped, instance, argv, kwargs):
        # Prep any incoming context headers
        metadata = argv[0].invocation_metadata
        metadata_dict = {}
        for c in metadata:
            metadata_dict[c.key] = c.value

        ctx = tracer.extract(opentracing.Format.BINARY, metadata_dict, disable_w3c_trace_context=True)

        with tracer.start_active_span("rpc-server", child_of=ctx) as scope:
            try:
                collect_tags(scope.span, instance, argv, kwargs)
                rv = wrapped(*argv, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    logger.debug("Instrumenting grpcio")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from __future__ import absolute_import

import wrapt
import re

from ....log import logger
from ....singletons import tracer
from .collectors import _storage_api

try:
    from google.cloud import storage

    logger.debug('Instrumenting google-cloud-storage')

    def _collect_tags(api_request):
        """
        Extract span tags from Google Cloud Storage API request. Returns None if the request is not
        supported.

        :param: dict
        :return: dict or None
        """
        method, path = api_request.get('method', None), api_request.get('path', None)

        if method not in _storage_api:
            return

        try:
            params = api_request.get('query_params', {})
            data = api_request.get('data', {})

            if path in _storage_api[method]:
                # check is any of string keys matches the path exactly
                return _storage_api[method][path](params, data)
            else:
                # look for a regex that matches the string
                for (matcher, collect) in _storage_api[method].items():
                    if not isinstance(matcher, re.Pattern):
                        continue

                    m = matcher.match(path)
                    if m is None:
                        continue

                    return collect(params, data, m)
        except Exception:
            logger.debug("instana.instrumentation.google.cloud.storage._collect_tags: ", exc_info=True)

    def execute_with_instana(wrapped, instance, args, kwargs):
        # batch requests are traced with finish_batch_with_instana()
        if isinstance(instance, storage.Batch):
            return wrapped(*args, **kwargs)

        parent_span = tracer.active_span

        # return early if we're not tracing
        if parent_span is None:
            return wrapped(*args, **kwargs)

        tags = _collect_tags(kwargs)

        # don't trace if the call is not instrumented
        if tags is None:
            logger.debug('uninstrumented Google Cloud Storage API request: %s' % kwargs)
            return wrapped(*args, **kwargs)

        with tracer.start_active_span('gcs', child_of=parent_span) as scope:
            for (k, v) in tags.items():
                scope.span.set_tag(k, v)

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return kv

    def download_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # return early if we're not tracing
        if parent_span is None:
            return wrapped(*args, **kwargs)

        with tracer.start_active_span('gcs', child_of=parent_span) as scope:
            scope.span.set_tag('gcs.op', 'objects.get')
            scope.span.set_tag('gcs.bucket', instance.bucket.name)
            scope.span.set_tag('gcs.object', instance.name)

            start = len(args) > 4 and args[4] or kwargs.get('start', None)
            if start is None:
                start = ''

            end = len(args) > 5 and args[5] or kwargs.get('end', None)
            if end is None:
                end = ''

            if start != '' or end != '':
                scope.span.set_tag('gcs.range', '-'.join((start, end)))

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return kv

    def upload_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # return early if we're not tracing
        if parent_span is None:
            return wrapped(*args, **kwargs)

        with tracer.start_active_span('gcs', child_of=parent_span) as scope:
            scope.span.set_tag('gcs.op', 'objects.insert')
            scope.span.set_tag('gcs.bucket', instance.bucket.name)
            scope.span.set_tag('gcs.object', instance.name)

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return kv

    def finish_batch_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # return early if we're not tracing
        if parent_span is None:
            return wrapped(*args, **kwargs)

        with tracer.start_active_span('gcs', child_of=parent_span) as scope:
            scope.span.set_tag('gcs.op', 'batch')
            scope.span.set_tag('gcs.projectId', instance._client.project)
            scope.span.set_tag('gcs.numberOfOperations', len(instance._requests))

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return kv

    wrapt.wrap_function_wrapper('google.cloud.storage._http', 'Connection.api_request', execute_with_instana)
    wrapt.wrap_function_wrapper('google.cloud.storage.blob', 'Blob._do_download', download_with_instana)
    wrapt.wrap_function_wrapper('google.cloud.storage.blob', 'Blob._do_upload', upload_with_instana)
    wrapt.wrap_function_wrapper('google.cloud.storage.batch', 'Batch.finish', finish_batch_with_instana)
except ImportError:
    pass
